
#include "../../waveOutPlayer.h"
#include "../../waveOutDriver.h"

#include "../ExtensibleFormat.h"

namespace Juna {
namespace Sound {

namespace {
bool LoadStream(void *dest,unsigned long destsize,const Format &format,
				I_InputStream &stream)
{
	if (format != stream.GetFormat())
		return false;
	unsigned long loadsize = stream.Read(dest,destsize);

	int fill = (format.bit == 8 && format.quantization == QT_UNSIGNED) ? 128 : 0;
	::memset(static_cast<char *>(dest) + loadsize,fill,destsize - loadsize);
	return true;
}
}//anonymous namespace

base_waveOutPoolPlayer::base_waveOutPoolPlayer(void)
  : hWaveOut(NULL),Buffer(0),
	PlayFormat(),
	PlayFlag(false),LoopFlag(false),VolumeFlag(false)
{
	::memset(&WaveHdr,0,sizeof(WAVEHDR));
	::memset(&LoopHdr,0,sizeof(WAVEHDR));
}

bool base_waveOutPoolPlayer::Start(void)
{
	if (LoopFlag)
		::waveOutWrite(hWaveOut,&LoopHdr,sizeof(WAVEHDR));
	else
		::waveOutWrite(hWaveOut,&WaveHdr,sizeof(WAVEHDR));

	return true;
}

bool base_waveOutPoolPlayer::Stop(void)
{
	::waveOutReset(hWaveOut);
	return true;
}

bool base_waveOutPoolPlayer::SetDeviceVolume(unsigned long iLeft,unsigned long iRight)
{
	iLeft = min(iLeft,0xFFFF);
	iRight = min(iRight,0xFFFF);
	DWORD setvol = iLeft | (iRight << 16);
	return ::waveOutSetVolume(hWaveOut,setvol) == MMSYSERR_NOERROR;
}

bool base_waveOutPoolPlayer::GetDeviceVolume(unsigned long &left,unsigned long &right) const
{
	DWORD getvol;
	if (::waveOutGetVolume(hWaveOut,&getvol) == MMSYSERR_NOERROR)
	{
		right = (getvol >> 16) & 0xFFFF;
		left = getvol & 0xFFFF;
		right += (right == 0xFFFF);
		left += (left == 0xFFFF);
		return true;
	}
	return false;
}

const I_PlayDriver *waveOutPoolPlayer::GetDriver(void)
{return &waveOutPlayDriver::GetInstance();}


void base_waveOutPoolPlayer::SetLoopFlag(bool loop)
{
	if (isPlaying())
	{
		if (!LoopFlag && loop)
			::waveOutWrite(hWaveOut,&LoopHdr,sizeof(WAVEHDR));
		else
			::waveOutBreakLoop(hWaveOut);
	}
	LoopFlag = loop;
}

long base_waveOutPoolPlayer::GetPlayPosition(void)
{
	MMRESULT mmr;
	MMTIME mmt;
	mmt.wType = TIME_SAMPLES;
	mmr = ::waveOutGetPosition(hWaveOut,&mmt,sizeof(mmt));
	if (mmr == MMSYSERR_NOERROR)
		return mmt.u.sample % (WaveHdr.dwBufferLength / PlayFormat.sample_size());
	return -1;
}



bool base_waveOutPoolPlayer::base_Initialize
	(int device,const Format &format,long bufsize)
{
	base_Terminalize();
	if (bufsize == 0)
		return false;

	const UINT *id = waveOutPlayDriver::GetInstance().GetWODeviceID(device);
	if (id == 0)
		return false;

	if (!OpenDevice(*id,format))
		return false;

//TvTCYɍ킹
	if (bufsize % format.sample_size())
		bufsize -= bufsize % format.sample_size();

	Buffer = static_cast<SharedBuffer *>(operator new(sizeof(SharedBuffer) + bufsize));
	if (!Buffer)
	{
		base_Terminalize();
		return false;
	}
	Buffer->Buffer = Buffer + 1;
	Buffer->ShareCount = 1;
	WaveHdr.lpData = LoopHdr.lpData = static_cast<LPSTR>(Buffer->Buffer);
	WaveHdr.dwBufferLength = LoopHdr.dwBufferLength = bufsize;
	WaveHdr.dwFlags = WHDR_DONE;
	LoopHdr.dwFlags = WHDR_DONE | WHDR_BEGINLOOP | WHDR_ENDLOOP;

	PlayFormat = format;

	return true;
}

bool base_waveOutPoolPlayer::OpenDevice
	(unsigned int deviceID,const Format &format)
{
	if ((format.bit == 8 && format.quantization == QT_SIGNED) ||
		(format.bit != 8 && format.quantization == QT_UNSIGNED))
		return false;

	WAVEFORMATEX wfx;
	SetWaveFormatEx(wfx,format);
	DWORD openflag = CALLBACK_NULL;
//	if (dd->ID == WAVE_MAPPER) openflag |= WAVE_MAPPED;
	HWAVEOUT hwo;
	if (format.bit > 16 || format.channel > 2 ||
		::waveOutOpen(&hwo,deviceID,&wfx,0,0,openflag) != MMSYSERR_NOERROR)
	{
		const int extsize = sizeof(WAVEFORMATEX) + 22;
		BYTE buf[extsize];
		WAVEFORMATEX *wfext = reinterpret_cast<WAVEFORMATEX *>(buf);
		memcpy(wfext,&wfx,sizeof(WAVEFORMATEX));
		wfext->cbSize = extsize - sizeof(WAVEFORMATEX);
		if (!MakeExtensibleFormat(wfext))
			return false;
		if (::waveOutOpen(&hwo,deviceID,wfext,0,0,openflag) != MMSYSERR_NOERROR)
			return false;
	}
	hWaveOut = hwo;
	return true;
}

bool base_waveOutPoolPlayer::base_Initialize(I_PoolPlayer &duplsrc)
{
	base_Terminalize();

	if (&waveOutPlayDriver::GetInstance() != duplsrc.GetDriver() &&
		&waveOutExVolumePlayDriver::GetInstance() != duplsrc.GetDriver())
		return false;

	base_waveOutPoolPlayer &bwop = static_cast<base_waveOutPoolPlayer &>(duplsrc);
	if (!bwop.IsInitialized())
		return false;

	if (!OpenDevice((UINT)bwop.hWaveOut,bwop.PlayFormat))
		return false;

	Buffer = bwop.Buffer;
	Buffer->ShareCount++;
	WaveHdr.lpData = LoopHdr.lpData = static_cast<LPSTR>(Buffer->Buffer);
	WaveHdr.dwBufferLength = LoopHdr.dwBufferLength = bwop.WaveHdr.dwBufferLength;
	WaveHdr.dwFlags = WHDR_DONE;
	LoopHdr.dwFlags = WHDR_DONE | WHDR_BEGINLOOP | WHDR_ENDLOOP;

	PlayFormat = bwop.PlayFormat;
	LoopFlag = bwop.LoopFlag;

	return true;
}


void base_waveOutPoolPlayer::base_Terminalize(void)
{
	if (Buffer)
	{
		::waveOutReset(hWaveOut);
		::waveOutUnprepareHeader(hWaveOut,&LoopHdr,sizeof(WAVEHDR));
		::waveOutUnprepareHeader(hWaveOut,&WaveHdr,sizeof(WAVEHDR));
		if (--Buffer->ShareCount <= 0)
			operator delete(Buffer);
		Buffer = 0;
		memset(&WaveHdr,0,sizeof(WaveHdr));
	}
	if (hWaveOut)
	{
		::waveOutClose(hWaveOut);
		hWaveOut = NULL;
	}
	PlayFormat.reset();
}

bool waveOutPoolPlayer::Initialize
	(int device,const Format &format,long iBufSize)
{
	if (!base_Initialize(device,format,iBufSize))
		return false;
	::waveOutPrepareHeader(hWaveOut,&WaveHdr,sizeof(WAVEHDR));
	::waveOutPrepareHeader(hWaveOut,&LoopHdr,sizeof(WAVEHDR));
	return true;
}
bool waveOutPoolPlayer::Initialize
	(int device,I_SeekableInputStream &stream)
{
	Format format = stream.GetFormat();
	long bufsize = stream.GetLength() * format.sample_size(); 
	if (!Initialize(device,format,bufsize))
		return false;
	if (stream.Seek(0) != 0 || !LoadStream(stream))
	{
		Terminalize();
		return false;
	}
	return true;
}
bool waveOutPoolPlayer::Initialize(I_PoolPlayer &iDuplicateSrc)
{
	if (!base_Initialize(iDuplicateSrc))
		return false;
	::waveOutPrepareHeader(hWaveOut,&WaveHdr,sizeof(WAVEHDR));
	::waveOutPrepareHeader(hWaveOut,&LoopHdr,sizeof(WAVEHDR));
	return true;
}

bool waveOutPoolPlayer::LoadStream(I_InputStream &stream)
{
	if (!Stop())
		return false;
	if (PlayFormat != stream.GetFormat())
		return false;

	::waveOutUnprepareHeader(hWaveOut,&LoopHdr,sizeof(WAVEHDR));
	::waveOutUnprepareHeader(hWaveOut,&WaveHdr,sizeof(WAVEHDR));

	unsigned long bufsize = GetBufferSize();
	unsigned long loadsize = stream.Read(Buffer->Buffer,bufsize);
	int fill = (PlayFormat.bit == 8) ? 128 : 0;
	::memset(static_cast<char *>(Buffer->Buffer) + loadsize,fill,bufsize - loadsize);

	::waveOutPrepareHeader(hWaveOut,&WaveHdr,sizeof(WAVEHDR));
	::waveOutPrepareHeader(hWaveOut,&LoopHdr,sizeof(WAVEHDR));

	return true;
}




const I_PlayDriver *waveOutExVolumePoolPlayer::GetDriver(void)
{return &waveOutExVolumePlayDriver::GetInstance();}

bool waveOutExVolumePoolPlayer::Initialize
	(int iDevice,const Format &iFormat,long iBufSize)
{
	if (!base_Initialize(iDevice,iFormat,iBufSize))
		return false;

	LPSTR data = static_cast<LPSTR>(operator new(WaveHdr.dwBufferLength));
	if (!data)
	{
		base_Terminalize();
		return false;
	}
	WaveHdr.lpData = LoopHdr.lpData = data;
	::waveOutPrepareHeader(hWaveOut,&WaveHdr,sizeof(WAVEHDR));
	::waveOutPrepareHeader(hWaveOut,&LoopHdr,sizeof(WAVEHDR));

	return false;
}
bool waveOutExVolumePoolPlayer::Initialize
	(int device,I_SeekableInputStream &stream)
{
	Format format = stream.GetFormat();
	long bufsize = stream.GetLength() * format.sample_size(); 
	if (!Initialize(device,format,bufsize))
		return false;
	if (stream.Seek(0) != 0 || !LoadStream(stream))
	{
		Terminalize();
		return false;
	}
	return true;
}
bool waveOutExVolumePoolPlayer::Initialize(I_PoolPlayer &iDuplicateSrc)
{
	if (!base_Initialize(iDuplicateSrc))
		return false;

	LPSTR data = static_cast<LPSTR>(operator new(WaveHdr.dwBufferLength));
	if (!data)
	{
		base_Terminalize();
		return false;
	}
	WaveHdr.lpData = LoopHdr.lpData = data;
	::waveOutPrepareHeader(hWaveOut,&WaveHdr,sizeof(WAVEHDR));
	::waveOutPrepareHeader(hWaveOut,&LoopHdr,sizeof(WAVEHDR));

	return false;
}

void waveOutExVolumePoolPlayer::Terminalize(void)
{
	LPSTR data = WaveHdr.lpData;
	base_Terminalize();
	if (data)
		operator delete(data);
}


bool waveOutExVolumePoolPlayer::SetBufferVolume
	(unsigned long left,unsigned long right)
{
	if (!IsInitialized())
		return false;
	LeftVolume = min(left,0x10000);
	RightVolume = min(right,0x10000);

	ChangeBufferVolume();	//Đł
	return true;
}

bool waveOutExVolumePoolPlayer::LoadStream(I_InputStream &stream)
{
	if (!Stop())
		return false;
	if (PlayFormat != stream.GetFormat())
		return false;

	unsigned long bufsize = GetBufferSize();
	unsigned long loadsize = stream.Read(Buffer->Buffer,bufsize);
	int fill = (PlayFormat.bit == 8) ? 128 : 0;
	::memset(static_cast<char *>(Buffer->Buffer) + loadsize,fill,bufsize - loadsize);

	::waveOutUnprepareHeader(hWaveOut,&WaveHdr,sizeof(WAVEHDR));
	::waveOutUnprepareHeader(hWaveOut,&LoopHdr,sizeof(WAVEHDR));

	ChangeBufferVolume();

	::waveOutPrepareHeader(hWaveOut,&WaveHdr,sizeof(WAVEHDR));
	::waveOutPrepareHeader(hWaveOut,&LoopHdr,sizeof(WAVEHDR));

	return true;
}

void waveOutExVolumePoolPlayer::ChangeBufferVolume(void)
{
//	unsigned long left = LeftVolume + (LeftVolume == 0xFFFF);
//	unsigned long right = RightVolume + (RightVolume == 0xFFFF);
	double left = (double)LeftVolume / (double)0x10000;
	double right = (double)RightVolume / (double)0x10000;
	long samples = WaveHdr.dwBufferLength / PlayFormat.sample_size();
	long i;
	switch (PlayFormat.bit)
	{
	case 8:
		{
		BYTE *dest = reinterpret_cast<BYTE *>(WaveHdr.lpData);
		BYTE *src = static_cast<BYTE *>(Buffer->Buffer);
		switch (PlayFormat.channel)
		{
		case 1:
			for (i = 0;i < samples;i++)
			{
				*dest = static_cast<BYTE>((static_cast<int>(*src) - 128) * left + 128);
				dest++;src++;
			}
			break;
		case 2:
			for (i = 0;i < samples;i++)
			{
				dest[0] = static_cast<BYTE>((static_cast<int>(src[0]) - 128) * left + 128);
				dest[1] = static_cast<BYTE>((static_cast<int>(src[1]) - 128) * right + 128);
				dest += 2;src += 2;
			}
			break;
		}
		break;
		}
	case 16:
		{
		short *dest = reinterpret_cast<short *>(WaveHdr.lpData);
		short *src = static_cast<short *>(Buffer->Buffer);
		switch (PlayFormat.channel)
		{
		case 1:
			for (i = 0;i < samples;i++)
			{
				*dest = static_cast<short>(*src * left);
				dest++;src++;
			}
			break;
		case 2:
			for (i = 0;i < samples;i++)
			{
				dest[0] = static_cast<short>(src[0] * left);
				dest[1] = static_cast<short>(src[1] * right);
				dest += 2;src += 2;
			}
			break;
		}
		break;
		}
	}
}

}//namespace Sound
}//namespace Juna

